from __future__ import annotations

from ctypes import _Pointer
from typing import Literal, cast

import jack_server._lib as lib
from jack_server._parameter import Parameter, get_params_from_jslist

SampleRate = Literal[44100, 48000]


class Driver:
    params: dict[str, Parameter]
    _ptr: _Pointer[lib.jackctl_driver_t]

    def __init__(self, ptr: _Pointer[lib.jackctl_driver_t]) -> None:
        self._ptr = ptr
        self._init_params()

    def _init_params(self) -> None:
        params_jslist = lib.jackctl_driver_get_parameters(self._ptr)
        self.params = get_params_from_jslist(params_jslist)

    @property
    def name(self) -> str:
        return cast(bytes, lib.jackctl_driver_get_name(self._ptr)).decode()

    @property
    def device(self) -> str:  # pragma: no cover
        return cast(bytes, self.params["device"].value).decode()

    @device.setter
    def device(self, __value: str) -> None:  # pragma: no cover
        self.params["device"].value = __value.encode()

    @property
    def rate(self) -> SampleRate:
        return cast(SampleRate, self.params["rate"].value)

    @rate.setter
    def rate(self, __value: SampleRate) -> None:
        self.params["rate"].value = __value

    @property
    def period(self) -> int:
        return cast(int, self.params["period"].value)

    @period.setter
    def period(self, __value: int) -> None:
        self.params["period"].value = __value

    def __repr__(self) -> str:
        return f"<jack_server.Driver name={self.name}>"



from ctypes import (
    CDLL,
    CFUNCTYPE,
    POINTER,
    Structure,
    Union,
    _Pointer,
    c_bool,
    c_char,
    c_char_p,
    c_int,
    c_uint,
    c_void_p,
)
from ctypes.util import find_library
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ctypes import _CData

_lib_names = ("libjackserver", "jackserver", "libjackserver64")


def get_library_name():
    for name in _lib_names:
        if result := find_library(name):
            return result

    raise RuntimeError("Couldn't find jackserver library")


lib = CDLL(get_library_name())


class JSList(Structure):
    data: "_CData"
    next: "_Pointer[JSList]"


JSList_p = POINTER(JSList)
JSList._fields_ = [("data", c_void_p), ("next", JSList_p)]


class jackctl_parameter_t(Structure):
    pass


jackctl_parameter_t_p = POINTER(jackctl_parameter_t)

jackctl_parameter_get_type = lib.jackctl_parameter_get_type
jackctl_parameter_get_type.argtypes = [jackctl_parameter_t_p]
jackctl_parameter_get_type.restype = c_uint

jackctl_parameter_get_name = lib.jackctl_parameter_get_name
jackctl_parameter_get_name.argtypes = [jackctl_parameter_t_p]
jackctl_parameter_get_name.restype = c_char_p


class jackctl_parameter_value(Union):
    _fields_ = [
        ("ui", c_uint),
        ("i", c_int),
        ("c", c_char),
        ("ss", c_char * 128),
        ("b", c_bool),
    ]

    i: int
    ui: int
    c: str
    ss: bytes
    b: bool


jackctl_parameter_value_p = POINTER(jackctl_parameter_value)

jackctl_parameter_set_value = lib.jackctl_parameter_set_value
jackctl_parameter_set_value.argtypes = [
    jackctl_parameter_t_p,
    jackctl_parameter_value_p,
]
jackctl_parameter_set_value.restype = c_bool

jackctl_parameter_get_value = lib.jackctl_parameter_get_value
jackctl_parameter_get_value.argtypes = [jackctl_parameter_t_p]
jackctl_parameter_get_value.restype = jackctl_parameter_value


class jackctl_driver_t(Structure):
    pass


jackctl_driver_t_p = POINTER(jackctl_driver_t)

jackctl_driver_get_parameters = lib.jackctl_driver_get_parameters
jackctl_driver_get_parameters.argtypes = [jackctl_driver_t_p]
jackctl_driver_get_parameters.restype = JSList_p

jackctl_driver_get_name = lib.jackctl_driver_get_name
jackctl_driver_get_name.argtypes = [jackctl_driver_t_p]
jackctl_driver_get_name.restype = c_char_p


class jackctl_server_t(Structure):
    pass


jackctl_server_t_p = POINTER(jackctl_server_t)

OnDeviceAcquire = CFUNCTYPE(c_bool, c_char_p)
OnDeviceRelease = CFUNCTYPE(None, c_char_p)
OnDeviceReservationLoop = CFUNCTYPE(None)

jackctl_server_create2 = lib.jackctl_server_create2
jackctl_server_create2.argtypes = [
    OnDeviceAcquire,
    OnDeviceRelease,
    OnDeviceReservationLoop,
]
jackctl_server_create2.restype = jackctl_server_t_p

jackctl_server_open = lib.jackctl_server_open
jackctl_server_open.argtypes = [jackctl_server_t_p, jackctl_driver_t_p]
jackctl_server_open.restype = c_bool

jackctl_server_start = lib.jackctl_server_start
jackctl_server_start.argtypes = [jackctl_server_t_p]
jackctl_server_start.restype = c_bool

jackctl_server_close = lib.jackctl_server_close
jackctl_server_close.argtypes = [jackctl_server_t_p]
jackctl_server_close.restype = c_bool

jackctl_server_stop = lib.jackctl_server_stop
jackctl_server_stop.argtypes = [jackctl_server_t_p]
jackctl_server_stop.restype = c_bool

jackctl_server_destroy = lib.jackctl_server_destroy
jackctl_server_destroy.argtypes = [jackctl_server_t_p]
jackctl_server_destroy.restype = c_void_p

jackctl_server_get_parameters = lib.jackctl_server_get_parameters
jackctl_server_get_parameters.argtypes = [jackctl_server_t_p]
jackctl_server_get_parameters.restype = JSList_p

jackctl_server_get_drivers_list = lib.jackctl_server_get_drivers_list
jackctl_server_get_drivers_list.argtypes = [jackctl_server_t_p]
jackctl_server_get_drivers_list.restype = JSList_p


PrintFunction = CFUNCTYPE(None, c_char_p)

jack_set_error_function = lib.jack_set_error_function
jack_set_error_function.argtypes = [PrintFunction]
jack_set_error_function.restype = None

jack_set_info_function = lib.jack_set_info_function
jack_set_info_function.argtypes = [PrintFunction]
jack_set_info_function.restype = None



from __future__ import annotations

from ctypes import _Pointer, cast
from typing import TYPE_CHECKING, Iterable, TypeVar

import jack_server._lib as lib

if TYPE_CHECKING:
    from ctypes import _CanCastTo

    _T = TypeVar("_T", bound=_CanCastTo)


def iterate_jslist(ptr: _Pointer[lib.JSList], type: type[_T]) -> Iterable[_T]:
    cur_ptr = ptr

    while cur_ptr:
        data = cur_ptr.contents.data
        cur_ptr = cur_ptr.contents.next
        yield cast(data, type)



from __future__ import annotations

from ctypes import _Pointer, pointer
from typing import Literal, Union, cast

import jack_server._lib as lib
from jack_server._jslist import iterate_jslist

ValueType = Union[int, str, bytes, bool]


class Parameter:
    _ptr: _Pointer[lib.jackctl_parameter_t]
    type: Literal[1, 2, 3, 4, 5]

    def __init__(self, ptr: _Pointer[lib.jackctl_parameter_t]) -> None:
        self._ptr = ptr
        self.type = lib.jackctl_parameter_get_type(self._ptr)

    @property
    def name(self) -> str:
        return cast(bytes, lib.jackctl_parameter_get_name(self._ptr)).decode()

    @property
    def value(self) -> ValueType:  # pragma: no cover
        val = cast(
            lib.jackctl_parameter_value, lib.jackctl_parameter_get_value(self._ptr)
        )

        if self.type == 1:
            # JackParamInt
            return val.i
        elif self.type == 2:
            # JackParamUInt
            return val.ui
        elif self.type == 3:
            # JackParamChar
            return val.c
        elif self.type == 4:
            # JackParamString
            return val.ss
        elif self.type == 5:
            # JackParamBool
            return val.b
        else:
            raise NotImplementedError

    @value.setter
    def value(self, val: ValueType) -> None:  # pragma: no cover
        val_obj = lib.jackctl_parameter_value()

        if self.type == 1:
            # JackParamInt
            val_obj.i = int(val)
        elif self.type == 2:
            # JackParamUInt
            val_obj.ui = int(val)
        elif self.type == 3:
            # JackParamChar
            assert isinstance(val, str) and len(val) == 1
            val_obj.c = val
        elif self.type == 4:
            # JackParamString
            assert isinstance(val, bytes)
            val_obj.ss = val
        elif self.type == 5:
            # JackParamBool
            val_obj.b = bool(val)
        else:
            raise NotImplementedError

        lib.jackctl_parameter_set_value(self._ptr, pointer(val_obj))

    def __repr__(self) -> str:
        return f"<jack_server.Parameter name={self.name!r} value={self.value!r}>"


def get_params_from_jslist(jslist: _Pointer[lib.JSList]) -> dict[str, Parameter]:
    params: dict[str, Parameter] = {}

    for ptr in iterate_jslist(jslist, lib.jackctl_parameter_t_p):
        param = Parameter(ptr)
        params[param.name] = param

    return params



from __future__ import annotations

from ctypes import _Pointer
from typing import Callable, cast

import jack_server._lib as lib
from jack_server._driver import Driver, SampleRate
from jack_server._jslist import iterate_jslist
from jack_server._parameter import Parameter, get_params_from_jslist


class JackServerError(RuntimeError):
    pass


class ServerNotStartedError(JackServerError):
    pass


class ServerNotOpenedError(JackServerError):
    pass


class DriverNotFoundError(JackServerError):
    pass


class SetByJack:
    pass


SetByJack_: SetByJack = SetByJack()


class Server:
    driver: Driver
    params: dict[str, Parameter]
    _ptr: _Pointer[lib.jackctl_server_t]
    _created: bool
    _opened: bool
    _started: bool
    _dont_garbage_collect: list[object]

    def __init__(
        self,
        *,
        name: str | SetByJack = SetByJack_,
        sync: bool | SetByJack = SetByJack_,
        realtime: bool | SetByJack = SetByJack_,  # TODO: Add docs
        driver: str,
        device: str | SetByJack = SetByJack_,
        rate: SampleRate | SetByJack = SetByJack_,
        period: int | SetByJack = SetByJack_,
    ) -> None:
        self._created = False
        self._opened = False
        self._started = False
        self._dont_garbage_collect = []

        self._create()
        self._init_params()
        self.driver = self._get_driver_by_name(driver)

        if not isinstance(name, SetByJack):
            self.name = name
        if not isinstance(sync, SetByJack):
            self.sync = sync
        if not isinstance(realtime, SetByJack):
            self.realtime = realtime
        if not isinstance(device, SetByJack):
            self.driver.device = device  # pragma: no cover (can't set driver on dummy driver that used in CI)
        if not isinstance(rate, SetByJack):
            self.driver.rate = rate
        if not isinstance(period, SetByJack):
            self.driver.period = period

    def _create(
        self,
        on_device_acquire: Callable[[bytes], bool] | None = None,
        on_device_release: Callable[[bytes], None] | None = None,
        on_device_reservation_loop: Callable[[], None] | None = None,
    ) -> None:
        if not on_device_acquire:
            on_device_acquire = lambda _: True
        if not on_device_release:
            on_device_release = lambda _: None
        if not on_device_reservation_loop:
            on_device_reservation_loop = lambda: None

        c_on_device_acquire = lib.OnDeviceAcquire(on_device_acquire)
        c_on_device_release = lib.OnDeviceRelease(on_device_release)
        c_on_device_reservation_loop = lib.OnDeviceReservationLoop(
            on_device_reservation_loop
        )

        args = (c_on_device_acquire, c_on_device_release, c_on_device_reservation_loop)
        self._dont_garbage_collect.extend(args)

        self._ptr = lib.jackctl_server_create2(*args)
        self._created = True

    def _open(self) -> None:
        self._opened = lib.jackctl_server_open(self._ptr, self.driver._ptr)
        if not self._opened:
            raise ServerNotOpenedError("Server couldn't be opened")

    def _start(self) -> None:
        self._started = lib.jackctl_server_start(self._ptr)
        if not self._started:
            raise ServerNotStartedError("Server couldn't be started")

    def _close(self) -> None:
        if self._opened:
            lib.jackctl_server_close(self._ptr)
            self._opened = False

    def _stop(self) -> None:
        if self._started:
            lib.jackctl_server_stop(self._ptr)
            self._started = False

    def _destroy(self) -> None:
        if self._created:
            lib.jackctl_server_destroy(self._ptr)
            self._created = False

    def start(self) -> None:
        self._open()
        self._start()

    def stop(self) -> None:
        self._stop()
        self._close()

    def __del__(self) -> None:
        self._destroy()

    def _init_params(self) -> None:
        jslist = lib.jackctl_server_get_parameters(self._ptr)
        self.params = get_params_from_jslist(jslist)

    def _get_driver_by_name(self, name: str) -> Driver:
        jslist = lib.jackctl_server_get_drivers_list(self._ptr)
        for ptr in iterate_jslist(jslist, lib.jackctl_driver_t_p):
            driver = Driver(ptr)
            if driver.name == name:
                return driver

        raise DriverNotFoundError(f"Driver not found: {name}")

    @property
    def name(self) -> str:
        return cast(bytes, self.params["name"].value).decode()

    @name.setter
    def name(self, __value: str) -> None:
        self.params["name"].value = __value.encode()

    @property
    def sync(self) -> bool:
        return cast(bool, self.params["sync"].value)

    @sync.setter
    def sync(self, __value: bool) -> None:
        self.params["sync"].value = __value

    @property
    def realtime(self) -> bool:
        return cast(bool, self.params["realtime"].value)

    @realtime.setter
    def realtime(self, __value: bool) -> None:
        self.params["realtime"].value = __value

    def __repr__(self) -> str:
        return f"<jack_server.Server driver={self.driver.name} started={self._started}>"



from __future__ import annotations

from functools import wraps
from typing import Callable

import jack_server._lib as lib

_dont_garbage_collect: list[object] = []


def _wrap_error_or_info_callback(
    callback: Callable[[str], None] | None
) -> lib.PrintFunction:
    if callback:

        @wraps(callback)
        def wrapped_callback(message: bytes) -> None:
            callback(message.decode())

    else:

        def wrapped_callback(message: bytes) -> None:
            pass

    c_callback = lib.PrintFunction(wrapped_callback)
    _dont_garbage_collect.append(c_callback)
    return c_callback


def set_info_function(callback: Callable[[str], None] | None) -> None:
    lib.jack_set_info_function(_wrap_error_or_info_callback(callback))


def set_error_function(callback: Callable[[str], None] | None) -> None:
    lib.jack_set_error_function(_wrap_error_or_info_callback(callback))



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['jack_server']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'jack-server',
    'version': '0.1.4',
    'description': 'Control JACK server with Python',
    'long_description': '[![Test](https://github.com/vrslev/jack_server/actions/workflows/test.yml/badge.svg)](https://github.com/vrslev/jack_server/actions/workflows/test.yml)\n\nControl [JACK](https://jackaudio.org/) audio server with Python.\nCan be used as replacement for [jackd](https://manpages.debian.org/buster/jackd2/jackd.1.en.html) for more robust configuration, for example, when using [`jack`](https://github.com/spatialaudio/jackclient-python) package.\n\n## Installation\n\n`pip install jack_server`\n\nAlso you need to have `jackserver` library on your machine, it comes with [JACK2](https://github.com/jackaudio/jack2). I had problems with apt-package on Ubuntu (`jackd2`), if you do too, compile jack yourself.\n\n## Usage\n\n### 🎛 `jack_server.Server`\n\nOn server creation you _can_ specify some parameters:\n\n```python\nimport jack_server\n\nserver = jack_server.Server(\n    name="myfancyserver",\n    sync=True,\n    realtime=False,\n    driver="coreaudio",\n    device="BuiltInSpeakerDevice",\n    rate=48000,\n    period=1024,\n)\nserver.start()\n\ninput()\n```\n\nThey are actually an equivalent of `jackd` flags:\n\n- `-n`, `--name` to `name`,\n- `-S`, `--sync` to `sync`,\n- `-R`, `--realtime`, `-r`, `--no-realtime` to `realtime`,\n- `-d` to `driver`,\n\nAnd driver arguments:\n\n- `-d`, `--device` to `device`,\n- `-r`, `--rate` to `rate`,\n- `-p`, `--period` to `period`,\n\n#### `start(self) -> None`\n\n_Open_ and _start_ the server. All state controlling methods are idempotent.\n\n#### `stop(self) -> None`\n\nStop and close server.\n\n#### `driver: jack_server.Driver`\n\nSelected driver.\n\n#### `name: str`\n\nActual server name. It is property that calls C code, so you can actually set the name.\n\n#### `sync: bool`\n\nWhether JACK runs in sync mode. Useful when you`re trying to send and receive multichannel audio.\n\n#### `realtime: bool`\n\nWhether JACK should start in realtime mode.\n\n#### `params: dict[str, jack_server.Parameter]`\n\nServer parameters mapped by name.\n\n### 💼 `jack_server.Driver`\n\nDriver (JACK backend), can be safely changed before server is started. Not supposed to be created by user code.\n\n#### `name: str`\n\nDriver name, read-only.\n\n#### `device: str`\n\nSelected device.\n\n#### `rate: jack_server.SampleRate`\n\nSampling rate.\n\n#### `period: int`\n\nBuffer size.\n\n#### `params: dict[str, jack_server.Parameter]`\n\nDriver parameters mapped by name.\n\n### 📻 `jack_server.SampleRate`\n\nValid sampling rate, `44100` or `48000`.\n\n### 🔻 `jack_server.Parameter`\n\nNot supposed to be created by user code.\n\n#### `name: str`\n\nRead-only verbose name of parameter.\n\n#### `value: int | str | bytes | bool`\n\nValue of the parameter, can be changed.\n\n### ❗️ `jack_server.set_info_function(callback: Callable[[str], None] | None) -> None`\n\nSet info output handler. By default JACK does is itself, i. e. output is being printed in stdout.\n\n### ‼️ `jack_server.set_error_function(callback: Callable[[str], None] | None) -> None`\n\nSet error output handler. By default JACK does is itself, i. e. output is being printed in stderr.\n',
    'author': 'Lev Vereshchagin',
    'author_email': 'mail@vrslev.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from jack_server._driver import Driver as Driver
from jack_server._driver import SampleRate as SampleRate
from jack_server._output import set_error_function as set_error_function
from jack_server._output import set_info_function as set_info_function
from jack_server._parameter import Parameter as Parameter
from jack_server._server import DriverNotFoundError as DriverNotFoundError
from jack_server._server import JackServerError as JackServerError
from jack_server._server import Server as Server
from jack_server._server import ServerNotOpenedError as ServerNotOpenedError
from jack_server._server import ServerNotStartedError as ServerNotStartedError
